# Numpy, sparse bib, matplotlib
import numpy as np
import scipy.sparse as sp  
import matplotlib.pyplot as plt
# fuer Test
from numpy.random import rand

""" Gram Matrix zu den quadratischen B-Splines zurueckgeben
    Args:
    nn -- 
    Returns:
    Gram Matrix mit n Hauptdiagonaleintraegen
"""
def get_gram(nn):
    return sp.diags([[1./120]*(nn-2),[13./60]*(nn-1),[11./20]*nn,[13./60]*(nn-1),[1./120]*(nn-2)], range(-2,3))/nn

""" CG-Verfahren zum LGS: AA xx = bb
    Args:
    AA -- Matrix
    bb -- rechte Seite
    tol -- angepeilte Norm des Residuums
    x0 -- Startiterierte
    Returns:
    xk -- finale Iterierte
"""
def cg(AA, bb, tol, x0=None):
    # Startiterierte 0
    if x0 == None:
        x0 = np.zeros_like(bb)
    # Iterationszaehler
    ii = 0
    nn = len(bb)
    xk = x0.copy()
    rk = bb - AA.dot(x0)
    dk = rk.copy()
    rsold = rk.dot(rk)
    err = np.sqrt(rsold)
    # Iterieren
    while ii < nn and err > tol:
        Ap = AA.dot(dk)
        rsold = rk.dot(rk)
        alphak = rsold / dk.dot(Ap)
        xk += alphak*dk
        rk -= alphak*Ap
        rsnew = rk.dot(rk)
        dk = rk+rsnew/rsold*dk
        # Norm des Residuums in Liste hinzufuegen
        err = np.sqrt(rsnew)
    # Rueckgabe
    return xk

""" Malt Kreise mit vorgebenen Mittelpunkten und vorgegebenen Radii
    Arguments:
    centers -- 2D Liste/array der Mittelpunkte, zuerst x, dann y
    radii -- Liste/array der Radii
    ax -- Zeichenflaeche, z.B mit plt.gca() oder fig.add_subplot
"""
def draw_circles(centers, radii, ax=plt.gca()):
    # Malflaeche so setzen, das alle Kreise sichtbar sind
    left = np.min(centers[0]-radii); width = np.max(centers[0]+radii)-left
    bottom = np.min(centers[1]-radii); height = np.max(centers[1]+radii)-bottom
    ax.set_xlim(left-0.05*width, left+1.05*width)
    ax.set_ylim(bottom-0.05*height, bottom+1.05*height)
    # Kreise malen
    circles = [None]*len(radii)
    for ii in xrange(len(radii)):
        circles[ii] = plt.Circle((centers[0,ii],centers[1,ii]), radii[ii], color='r', fill=False, clip_on=False)
        ax.add_artist(circles[ii])
    return ax

####################################################################################################

""" Malt zur Matrix AA ein approximiertes Bendixson Rechteck mit Abschaetzung nach Gerschgorin
    Arguments:
    AA - Matrix
"""
def plot_bendixson(AA, ax=None):
    if ax == None:
        ax = plt.gca()
    # A0 Abschaetzungen
    A0 = 0.5*(AA+np.conjugate(AA.T))
                      
                                                                                                                 
    # A1 Abschaetzungen, komplexe Zahlen in Python werden mit j deklariert, hier also floating Point "2i"
    A1 =                        /(2.j)
                      
                                                                                                                 
    # Rechteck bestimmen
    left =              
    right =              
    bottom =              
    top =              
    # Rechteck malen
    ax.plot([left, right, right, left, left], [bottom, bottom, top, top, bottom], 'g-')

""" Malt zur reelen Matrix AA die zugehoerigen Gerschgorin Kreise und Bendixson Rechtecke
    Arguments:
    AA -- Matrix
"""
def plot_gerschgorin(AA, ax=None):
    if ax == None:
        ax = plt.gca()
    # Radii der Kreise bestimmen
                   
                                  
    radii =                                                  
    # Mittelpunkte
    centers =              
    # Falls, Diagonalelemente identisch sind, Radii zusammenfassen
    filtered_centers, inverse = np.unique(centers, return_inverse=True)
    filtered_radii = np.array([np.max(radii[np.where(inverse==ii)]) for ii in xrange(len(filtered_centers))])
    # Mittelpunkte malen, Hinweis: real, imag fuer Real, bzw Imaginaerteil
    ax.plot(                                                                            )
    # Und noch Striche von Mittelpunkt bis Kreis um zu sehen, welcher Kreis zu welchem Mittelpunkt gehoert
    ax.plot([                         ,                          ], [                         ,                                         ], 'k:')
    # Kreise malen
    draw_circles(np.vstack((                         ,                          )),               , ax)
    # Bild zeigen

""" Zu Matrix AA die Eigenwerteinschliessungen malen
    Arguments:
    AA -- Matrix
"""
def plot_einschliessungen(AA, fig=None):
    if fig == None:
        fig = plt.figure()
    ax = fig.add_subplot(111)
    plot_gerschgorin(AA, ax)
    plot_bendixson(AA, ax)
                                  
                                             
    fig.show()

""" Vektoriteration nach von Mises
    Arguments:
    AA -- Matrix
    eps -- Abbruchkriterium
    x0 -- Startvektor
"""
def vektoriteration(AA, eps, x0=None):
    if x0 == None:
        new = np.eye(np.shape(AA)[1])[np.argmax(np.abs(AA).diagonal())]
    else:
        new = x0.copy()
    iterations = 0
    xs = []
    lambdas = []
    #Schleife
    while True:
        iterations += 1
                                   
                                                             
        # Iterierte speichern
                   
        # Eigenwertapproximation speichern
                                                          
        # Abbruch
        if ((old-new).dot(old-new) < eps*eps):
            break
    return iterations, xs, lambdas

""" Inverse Vektoriteration
    Arguments
    AA -- Matrix
    mu -- Shift
    eps -- Abbruchkriterium
    x0 -- Startvektor
"""
def inverse_vektoriteration(AA, mu, eps, x0=None):
    if x0 == None:
        new = np.eye(np.shape(AA)[1])[np.argmax(np.abs(AA).diagonal())]
    else:
        new = x0.copy()
    iterations = 0
    xs = []
    lambdas = []
    BB = AA-mu*sp.eye(np.shape(AA)[0])
    #Schleife
    while True:
        iterations += 1
                 = cg(BB,              )     
                                                             
        # Iterierte speichern
                   
        # Eigenwertapproximations speichern
                                                          
        # Abbruch
        if ((old-new).dot(old-new) < eps*eps):
            break
    return iterations, xs, lambdas

""" Konvergenzbild zum dominanten Eigenwert malen
    Arguments:
    AA -- Matrix
    eps -- Abbruchkriterium
    x0 -- Startvektor
    mu -- Shift, falls gegeben, benutze inverse Iteration
"""
def plot_vektoriteration(AA, eps, x0=None, mu=None):
    # Rechnen
    iterations, xs, lambdas = vektoriteration(AA, eps, x0) if mu == None else inverse_vektoriteration(AA, mu, eps, x0)
                        
    # Zur Probe die Eigenwertapproximations und zwei Einschliessungen ausgeben
    idx = np.where(np.abs(xs[-1]) > 0)
    xfil = xs[-1][idx]
    print np.sqrt(bla.dot(bla)/(xs[-1].dot(xs[-1]))), np.min(bla[idx]/xfil), np.max(bla[idx]/xfil)
    # Bild
    fig = plt.figure()
    ax = fig.add_subplot(121)
    # Eigenwertiterierte malen, also die Approximationen an den dominanten Eigenwert
                                                               
                                                                 
                                                     
    # Beschriftungen
                                  
                                             
    # Fehler
    ax = fig.add_subplot(122)
    ax.semilogy(                                                                           )
    ax.set_xlabel(r'Iterationen')
    ax.set_ylabel(r'Fehler zu letzter Eigenwertapproximation')
    fig.show()

if __name__ == '__main__':
    eps = 1e-7
    # Test
    AA = np.array([[4, 0, 3], [0, -1, 1], [1, 1, 0]])
    plot_einschliessungen(AA)
    plt.gcf().savefig('test_0.pdf')
    BB = np.diag(rand(10))+(1j)*np.diag(rand(10))+np.diag(rand(9), 1)
    BB = BB+BB.T
    plot_einschliessungen(BB)
    plt.gcf().savefig('test_1.pdf')
    # Aufgabe
    for ii in xrange(10, 21, 10):
        AA = get_gram(ii)
        plot_einschliessungen(AA)
        plt.gcf().savefig(str(ii)+'_einschliessungen.pdf')
        plot_vektoriteration(AA, eps)
        plt.gcf().savefig(str(ii)+'_iteration.pdf')
        plot_vektoriteration(AA, eps, mu=.1)
        plt.gcf().savefig(str(ii)+'_inverse_iteration.pdf')
